@@ -10,6 +10,7 @@
 #include "list.h"
 #include "xmalloc.h"
 #include "cgroup.h"
+#include "cgroup-props.h"
 #include "cr_options.h"
 #include "pstree.h"
 #include "proc_parse.h"
@@ -24,73 +25,6 @@
 #include "images/core.pb-c.h"
 #include "images/cgroup.pb-c.h"
 
-/*
- * These string arrays have the names of all the properties that will be
- * restored. To add a property for a cgroup type, add it to the
- * corresponding char array above the NULL terminator. If you are adding
- * a new cgroup family all together, you must also edit get_known_properties()
- * Currently the code only supports properties with 1 value
- */
-
-static const char *cpu_props[] = {
-	"cpu.shares",
-	"cpu.cfs_period_us",
-	"cpu.cfs_quota_us",
-	"cpu.rt_period_us",
-	"cpu.rt_runtime_us",
-	"notify_on_release",
-	NULL
-};
-
-static const char *memory_props[] = {
-	/* limit_in_bytes and memsw.limit_in_bytes must be set in this order */
-	"memory.limit_in_bytes",
-	"memory.memsw.limit_in_bytes",
-	"memory.use_hierarchy",
-	"notify_on_release",
-	NULL
-};
-
-static const char *cpuset_props[] = {
-	/*
-	 * cpuset.cpus and cpuset.mems must be set before the process moves
-	 * into its cgroup; they are "initialized" below to whatever the root
-	 * values are in copy_special_cg_props so as not to cause ENOSPC when
-	 * values are restored via this code.
-	 */
-	"cpuset.cpus",
-	"cpuset.mems",
-	"cpuset.memory_migrate",
-	"cpuset.cpu_exclusive",
-	"cpuset.mem_exclusive",
-	"cpuset.mem_hardwall",
-	"cpuset.memory_spread_page",
-	"cpuset.memory_spread_slab",
-	"cpuset.sched_load_balance",
-	"cpuset.sched_relax_domain_level",
-	"notify_on_release",
-	NULL
-};
-
-static const char *blkio_props[] = {
-	"blkio.weight",
-	"notify_on_release",
-	NULL
-};
-
-static const char *freezer_props[] = {
-	"notify_on_release",
-	NULL
-};
-
-static const char *global_props[] = {
-	"cgroup.clone_children",
-	"notify_on_release",
-	"cgroup.procs",
-	"tasks",
-	NULL
-};
-
 /*
  * This structure describes set of controller groups
  * a task lives in. The cg_ctl entries are stored in
@@ -190,7 +124,7 @@ static struct cg_set *get_cg_set(struct list_head *ctls, unsigned int n_ctls, bo
 	if (cs) {
 		cs->id = cg_set_ids++;
 		INIT_LIST_HEAD(&cs->ctls);
-		list_splice(ctls, &cs->ctls);
+		list_splice_init(ctls, &cs->ctls);
 		cs->n_ctls = n_ctls;
 		list_add_tail(&cs->l, &cg_sets);
 		n_sets++;
@@ -203,7 +137,9 @@ static struct cg_set *get_cg_set(struct list_head *ctls, unsigned int n_ctls, bo
 		}
 
 		if (collect && collect_cgroups(&cs->ctls)) {
-			put_ctls(ctls);
+			list_del(&cs->l);
+			n_sets--;
+			put_ctls(&cs->ctls);
 			xfree(cs);
 			return NULL;
 		}
@@ -419,33 +355,14 @@ static void free_all_cgroup_props(struct cgroup_dir *ncd)
 	ncd->n_properties = 0;
 }
 
-static const char **get_known_properties(char *controller)
-{
-	const char **prop_arr = NULL;
-
-	if (!strcmp(controller, "cpu"))
-		prop_arr = cpu_props;
-	else if (!strcmp(controller, "memory"))
-		prop_arr = memory_props;
-	else if (!strcmp(controller, "cpuset"))
-		prop_arr = cpuset_props;
-	else if (!strcmp(controller, "blkio"))
-		prop_arr = blkio_props;
-	else if (!strcmp(controller, "freezer"))
-		prop_arr = freezer_props;
-
-	return prop_arr;
-}
-
-static int dump_cg_props_array(const char *fpath, struct cgroup_dir *ncd,
-			       const char **prop_arr)
+static int dump_cg_props_array(const char *fpath, struct cgroup_dir *ncd, const cgp_t *cgp)
 {
 	int j;
 	char buf[PATH_MAX];
 	struct cgroup_prop *prop;
 
-	for (j = 0; prop_arr != NULL && prop_arr[j] != NULL; ++j) {
-		if (snprintf(buf, PATH_MAX, "%s/%s", fpath, prop_arr[j]) >= PATH_MAX) {
+	for (j = 0; cgp && j < cgp->nr_props; j++) {
+		if (snprintf(buf, PATH_MAX, "%s/%s", fpath, cgp->props[j]) >= PATH_MAX) {
 			pr_err("snprintf output was truncated\n");
 			return -1;
 		}
@@ -455,7 +372,7 @@ static int dump_cg_props_array(const char *fpath, struct cgroup_dir *ncd,
 			continue;
 		}
 
-		prop = create_cgroup_prop(prop_arr[j]);
+		prop = create_cgroup_prop(cgp->props[j]);
 		if (!prop) {
 			free_all_cgroup_props(ncd);
 			return -1;
@@ -467,6 +384,28 @@ static int dump_cg_props_array(const char *fpath, struct cgroup_dir *ncd,
 			return -1;
 		}
 
+		if (!strcmp("memory.oom_control", cgp->props[j])) {
+			char *new;
+			int disable;
+
+			if (sscanf(prop->value, "oom_kill_disable %d\n", &disable) != 1) {
+				pr_err("couldn't scan oom state from %s\n", prop->value);
+				free_cgroup_prop(prop);
+				free_all_cgroup_props(ncd);
+				return -1;
+			}
+
+			if (asprintf(&new, "%d", disable) < 0) {
+				pr_err("couldn't aloocate new oom value\n");
+				free_cgroup_prop(prop);
+				free_all_cgroup_props(ncd);
+				return -1;
+			}
+
+			xfree(prop->value);
+			prop->value = new;
+		}
+
 		pr_info("Dumping value %s from %s/%s\n", prop->value, fpath, prop->name);
 		list_add_tail(&prop->list, &ncd->properties);
 		ncd->n_properties++;
@@ -481,15 +420,14 @@ static int add_cgroup_properties(const char *fpath, struct cgroup_dir *ncd,
 	int i;
 
 	for (i = 0; i < controller->n_controllers; ++i) {
+		const cgp_t *cgp = cgp_get_props(controller->controllers[i]);
 
-		const char **prop_arr = get_known_properties(controller->controllers[i]);
-
-		if (dump_cg_props_array(fpath, ncd, prop_arr) < 0) {
+		if (dump_cg_props_array(fpath, ncd, cgp) < 0) {
 			pr_err("dumping known properties failed");
 			return -1;
 		}
 
-		if (dump_cg_props_array(fpath, ncd, global_props) < 0) {
+		if (dump_cg_props_array(fpath, ncd, &cgp_global) < 0) {
 			pr_err("dumping global properties failed");
 			return -1;
 		}
@@ -549,8 +487,14 @@ static int add_cgroup(const char *fpath, const struct stat *sb, int typeflag)
 
 		INIT_LIST_HEAD(&ncd->properties);
 		ncd->n_properties = 0;
-		if (add_cgroup_properties(fpath, ncd, current_controller) < 0)
+		if (add_cgroup_properties(fpath, ncd, current_controller) < 0) {
+			list_del(&ncd->siblings);
+			if (mtype == PARENT_MATCH)
+				match->n_children--;
+			else if (mtype == NO_MATCH)
+				current_controller->n_heads--;
 			goto out;
+		}
 	}
 
 	return 0;
@@ -603,7 +547,7 @@ static int collect_cgroups(struct list_head *ctls)
 
 	list_for_each_entry(cc, ctls, l) {
 		char path[PATH_MAX], mopts[1024];
-		char *name, prefix[] = ".criu.cgmounts.XXXXXX";
+		char prefix[] = ".criu.cgmounts.XXXXXX";
 		struct cg_controller *cg;
 
 		current_controller = NULL;
@@ -634,13 +578,10 @@ static int collect_cgroups(struct list_head *ctls)
 		if (!opts.manage_cgroups)
 			continue;
 
-		if (strstartswith(cc->name, "name=")) {
-			name = cc->name + 5;
+		if (strstartswith(cc->name, "name="))
 			snprintf(mopts, sizeof(mopts), "none,%s", cc->name);
-		} else {
-			name = cc->name;
-			snprintf(mopts, sizeof(mopts), "%s", name);
-		}
+		else
+			snprintf(mopts, sizeof(mopts), "%s", cc->name);
 
 		if (mkdtemp(prefix) == NULL) {
 			pr_perror("can't make dir for cg mounts");
@@ -867,12 +808,21 @@ static int dump_controllers(CgroupEntry *cg)
 	return 0;
 }
 
+static void free_sets(CgroupEntry *cg, unsigned nr)
+{
+	unsigned i;
+
+	for (i = 0; i < nr; i++)
+		xfree(cg->sets[i]->ctls);
+	xfree(cg->sets);
+}
+
 
 static int dump_sets(CgroupEntry *cg)
 {
 	struct cg_set *set;
 	struct cg_ctl *ctl;
-	int s, c;
+	unsigned s, c;
 	void *m;
 	CgSetEntry *se;
 	CgMemberEntry *ce;
@@ -902,8 +852,10 @@ static int dump_sets(CgroupEntry *cg)
 		m = xmalloc(se->n_ctls * (sizeof(CgMemberEntry *) + sizeof(CgMemberEntry)));
 		se->ctls = m;
 		ce = m + se->n_ctls * sizeof(CgMemberEntry *);
-		if (!m)
+		if (!m) {
+			free_sets(cg, s);
 			return -1;
+		}
 
 		c = 0;
 		list_for_each_entry(ctl, &set->ctls, l) {
@@ -927,12 +879,16 @@ static int dump_sets(CgroupEntry *cg)
 int dump_cgroups(void)
 {
 	CgroupEntry cg = CGROUP_ENTRY__INIT;
+	int ret = -1;
 
 	BUG_ON(!criu_cgset || !root_cgset);
 
 	/*
 	 * Check whether root task lives in its own set as compared
-	 * to criu. If yes, we should not dump anything.
+	 * to criu. If yes, we should not dump anything. Note that
+	 * list_is_singular() is slightly wrong here: if the criu cgset has
+	 * empty cgroups, those will not be restored on the target host, since
+	 * we're not dumping anything here.
 	 */
 
 	if (root_cgset == criu_cgset && list_is_singular(&cg_sets)) {
@@ -942,11 +898,16 @@ int dump_cgroups(void)
 
 	if (dump_sets(&cg))
 		return -1;
-	if (dump_controllers(&cg))
-		return -1;
+	if (dump_controllers(&cg)) {
+		goto err;
+	}
 
 	pr_info("Writing CG image\n");
-	return pb_write_one(img_from_set(glob_imgset, CR_FD_CGROUP), &cg, PB_CGROUP);
+	ret = pb_write_one(img_from_set(glob_imgset, CR_FD_CGROUP), &cg, PB_CGROUP);
+err:
+	free_sets(&cg, cg.n_sets);
+	xfree(cg.controllers);
+	return ret;
 }
 
 static int ctrl_dir_and_opt(CgControllerEntry *ctl, char *dir, int ds,
@@ -980,9 +941,15 @@ static int ctrl_dir_and_opt(CgControllerEntry *ctl, char *dir, int ds,
 	return doff;
 }
 
-static const char *special_cpuset_props[] = {
+/* Some properties cannot be restored after the cgroup has children or tasks in
+ * it. We restore these properties as soon as the cgroup is created.
+ */
+static const char *special_props[] = {
 	"cpuset.cpus",
 	"cpuset.mems",
+	"memory.kmem.limit_in_bytes",
+	"memory.swappiness",
+	"memory.use_hierarchy",
 	NULL,
 };
 
@@ -1206,7 +1173,7 @@ static int restore_perms(int fd, const char *path, CgroupPerms *perms)
 static int restore_cgroup_prop(const CgroupPropEntry * cg_prop_entry_p,
 			       char *path, int off)
 {
-	int cg, fd, len, ret = -1;;
+	int cg, fd, len, ret = -1;
 	CgroupPerms *perms = cg_prop_entry_p->perms;
 
 	if (!cg_prop_entry_p->value) {
@@ -1224,7 +1191,7 @@ static int restore_cgroup_prop(const CgroupPropEntry * cg_prop_entry_p,
 	cg = get_service_fd(CGROUP_YARD);
 	fd = openat(cg, path, O_WRONLY);
 	if (fd < 0) {
-		pr_err("bad file stream?");
+		pr_perror("bad cgroup path: %s", path);
 		return -1;
 	}
 
@@ -1232,20 +1199,22 @@ static int restore_cgroup_prop(const CgroupPropEntry * cg_prop_entry_p,
 		goto out;
 
 	/* skip these two since restoring their values doesn't make sense */
-	if (!strcmp(cg_prop_entry_p->name, "cgroup.procs") || !strcmp(cg_prop_entry_p->name, "tasks"))
+	if (!strcmp(cg_prop_entry_p->name, "cgroup.procs") || !strcmp(cg_prop_entry_p->name, "tasks")) {
+		ret = 0;
 		goto out;
+	}
 
 	len = strlen(cg_prop_entry_p->value);
 	if (write(fd, cg_prop_entry_p->value, len) != len) {
-		pr_perror("Failed writing %s to %s\n", cg_prop_entry_p->value, path);
+		pr_perror("Failed writing %s to %s", cg_prop_entry_p->value, path);
 		goto out;
 	}
 
+	ret = 0;
+
 out:
 	if (close(fd) != 0)
 		pr_perror("Failed closing %s", path);
-	else
-		ret = 0;
 
 	return ret;
 }
@@ -1293,6 +1262,25 @@ static void add_freezer_state_for_restore(CgroupPropEntry *entry, char *path, si
 	freezer_path[path_len] = 0;
 }
 
+static int next_device_entry(char *buf)
+{
+	char *pos = buf;
+
+	while (1) {
+		if (*pos == '\n') {
+			*pos = '\0';
+			pos++;
+			break;
+		} else if (*pos == '\0') {
+			break;
+		}
+
+		pos++;
+	}
+
+	return pos - buf;
+}
+
 static int prepare_cgroup_dir_properties(char *path, int off, CgroupDirEntry **ents,
 					 unsigned int n_ents)
 {
@@ -1321,8 +1309,8 @@ static int prepare_cgroup_dir_properties(char *path, int off, CgroupDirEntry **e
 				 * the restore to fail if some other task has
 				 * entered the cgroup.
 				 */
-				for (k = 0; special_cpuset_props[k]; k++) {
-					if (!strcmp(e->properties[j]->name, special_cpuset_props[k])) {
+				for (k = 0; special_props[k]; k++) {
+					if (!strcmp(e->properties[j]->name, special_props[k])) {
 						special = true;
 						break;
 					}
@@ -1331,8 +1319,67 @@ static int prepare_cgroup_dir_properties(char *path, int off, CgroupDirEntry **e
 				if (special)
 					continue;
 
-				if (restore_cgroup_prop(e->properties[j], path, off2) < 0)
+				/* The devices cgroup must be restored in a
+				 * special way: only the contents of
+				 * devices.list can be read, and it is a
+				 * whitelist of all the devices the cgroup is
+				 * allowed to create. To re-creat this
+				 * whitelist, we first deny everything via
+				 * devices.deny, and then write the list back
+				 * into devices.allow.
+				 *
+				 * Further, we must have a write() call for
+				 * each line, because the kernel only parses
+				 * the first line of any write().
+				 */
+				if (!strcmp(e->properties[j]->name, "devices.list")) {
+					CgroupPropEntry *pe = e->properties[j];
+					char *old_val = pe->value, *old_name = pe->name;
+					int ret;
+					char *pos;
+
+					/* A bit of a fudge here. These are
+					 * write only by owner by default, but
+					 * the container engine could have
+					 * changed the perms. We should come up
+					 * with a better way to restore all of
+					 * this stuff.
+					 */
+					pe->perms->mode = 0200;
+
+					pe->name = "devices.deny";
+					pe->value = "a";
+					ret = restore_cgroup_prop(e->properties[j], path, off2);
+					pe->name = old_name;
+					pe->value = old_val;
+
+					if (ret < 0)
+						return -1;
+
+					pe->name = xstrdup("devices.allow");
+					if (!pe->name) {
+						pe->name = old_name;
+						return -1;
+					}
+					xfree(old_name);
+
+					pos = pe->value;
+					while (*pos) {
+						int offset = next_device_entry(pos);
+						pe->value = pos;
+						ret = restore_cgroup_prop(pe, path, off2);
+						if (ret < 0) {
+							pe->value = old_val;
+							return -1;
+						}
+						pos += offset;
+					}
+					pe->value = old_val;
+
+				} else if (restore_cgroup_prop(e->properties[j], path, off2) < 0) {
 					return -1;
+				}
+
 			}
 		}
 skip:
@@ -1364,21 +1411,30 @@ int prepare_cgroup_properties(void)
 	return 0;
 }
 
-static int restore_special_cpuset_props(char *paux, size_t off, CgroupDirEntry *e)
+static int restore_special_props(char *paux, size_t off, CgroupDirEntry *e)
 {
 	int i, j;
 
-	pr_info("Restore special cpuset props\n");
+	pr_info("Restore special props\n");
 
-	for (i = 0; special_cpuset_props[i]; i++) {
-		const char *name = special_cpuset_props[i];
+	for (i = 0; special_props[i]; i++) {
+		const char *name = special_props[i];
 
 		for (j = 0; j < e->n_properties; j++) {
 			CgroupPropEntry *prop = e->properties[j];
 
-			if (strcmp(name, prop->name) == 0)
-				if (restore_cgroup_prop(prop, paux, off) < 0)
+			if (strcmp(name, prop->name) == 0) {
+				/* XXX: we can drop this hack and make
+				 * memory.swappiness a regular property when we
+				 * drop support for kernels < 3.16. See 3dae7fec5.
+				 */
+				if (!strcmp(prop->name, "memory.swappiness") &&
+						!strcmp(prop->value, "60")) {
+					continue;
+				} else if (restore_cgroup_prop(prop, paux, off) < 0) {
 					return -1;
+				}
+			}
 		}
 	}
 
@@ -1434,8 +1490,8 @@ static int prepare_cgroup_dirs(char **controllers, int n_controllers, char *paux
 				return -1;
 
 			for (j = 0; j < n_controllers; j++) {
-				if (strcmp(controllers[j], "cpuset") == 0) {
-					if (restore_special_cpuset_props(paux, off2, e) < 0) {
+				if (!strcmp(controllers[j], "cpuset") || !strcmp(controllers[j], "memory")) {
+					if (restore_special_props(paux, off2, e) < 0) {
 						pr_err("Restoring special cpuset props failed!\n");
 						return -1;
 					}
@@ -1575,66 +1631,88 @@ err:
 }
 
 static int rewrite_cgsets(CgroupEntry *cge, char **controllers, int n_controllers,
-			  char **from, char *to)
+			  char **dir_name, char *newroot)
 {
-	int i, j;
-	bool set_from = false;
+	size_t dirlen = strlen(*dir_name);
+	char *dir = *dir_name;
+	char *dirnew = NULL;
+	size_t i, j;
+
+	/*
+	 * For example we may have the following in the image:
+	 *
+	 * set
+	 * 	name "hugetlb"
+	 * 	path "/300"
+	 *
+	 * controller
+	 * 	cnames hugetlb
+	 * 	dirs
+	 * 		dirname "300"
+	 * 		properties ...
+	 *
+	 * when we're switching to a new root we need to change
+	 * @path and don't forget to update the @dirname into
+	 * new state.
+	 */
 
 	for (i = 0; i < cge->n_sets; i++) {
 		CgSetEntry *set = cge->sets[i];
+
 		for (j = 0; j < set->n_ctls; j++) {
 			CgMemberEntry *cg = set->ctls[j];
-			char *tmp = cg->path, *tmp2 = NULL;
 
+			/*
+			 * Make sure if it's same controller
+			 * and its path with stripping leading
+			 * "/" is matching to be renamed.
+			 */
 			if (!(cgroup_contains(controllers, n_controllers, cg->name) &&
-					/* +1 to get rid of leading / */
-					strstartswith(cg->path + 1, *from)))
+					strstartswith(cg->path + 1, dir)))
 				continue;
 
-			/* If this cgset has a cgns prefix, let's use
-			 * that as the start of the root replacement.
-			 */
 			if (cg->has_cgns_prefix && cg->cgns_prefix) {
-				/* Rewrite the group dir to match the
-				 * prefix. We can do this exactly once
-				 * since we know all the tasks are in
-				 * the same cgroup ns (and thus have
-				 * the same per-controller prefix path)
-				 * since we don't support nesting.
-				 */
-				if (!set_from) {
-					set_from = true;
-					/* -1 because cgns_prefix includes leading / */
-					*from = xsprintf("%s%s", to, (*from) + cg->cgns_prefix - 1);
-				}
+				char *prev = cg->path;
 
-				cg->path = xsprintf("%s%s", to, cg->path +
-							cg->cgns_prefix);
-				cg->cgns_prefix = strlen(to);
+				cg->path = xsprintf("%s%s", newroot, cg->path + cg->cgns_prefix);
+				if (!cg->path) {
+					cg->path = prev;
+					return -ENOMEM;
+				}
+				xfree(prev);
+				cg->cgns_prefix = strlen(newroot);
+
+				if (!dirnew) {
+					/* -1 because cgns_prefix includes leading "/" */
+					dirnew = xsprintf("%s%s", newroot, dir + cg->cgns_prefix - 1);
+					if (!dirnew)
+						return -ENOMEM;
+				}
 			} else {
-				/* otherwise, use the old rewriting strategy */
-				cg->path = xsprintf("%s%s", to, cg->path +
-							strlen(*from) + 1);
-				if (!set_from) {
-					set_from = true;
-					*from = xstrdup(to);
+				char *prev = cg->path;
+				/*
+				 * If no prefix present simply rename the
+				 * root but make sure the rest of path is
+				 * untouched.
+				 */
+				cg->path = xsprintf("%s%s", newroot,
+						    cg->path + dirlen + 1);
+				if (!cg->path) {
+					cg->path = prev;
+					return -ENOMEM;
+				}
+				xfree(prev);
+				if (!dirnew) {
+					dirnew = xstrdup(newroot);
+					if (!dirnew)
+						return -ENOMEM;
 				}
 			}
-
-			if (tmp2) {
-				if (!*from)
-					return -1;
-
-				xfree(tmp2);
-			}
-
-			if (!cg->path)
-				return -1;
-
-			free(tmp);
 		}
-
 	}
+
+	xfree(dir);
+	*dir_name = dirnew;
 	return 0;
 }
 
